#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "gpsdk/gpsdk.h"
#include "player.h"
#include "fformat.h"
#include "myendian.h"

#include "gfx/gfx.h"

#include "gfx/bgpicraw.h"
#include "gfx/bgpicpal.h"

#include "gfx/fontpal.h"

#include "gfx/sintab.h"

u8 * screenbuf;
GPDRAWSURFACE gpDraw[2];

/* Converts 8bit rgb values to a GP32 palette value */
#define GP_RGB24(r,g,b) (((((r>>3))&0x1f)<<11)|((((g>>3))&0x1f)<<6)|((((b>>3))&0x1f)<<1))

sidTune * myTune = NULL;
emuEngine * myEmuEngine = NULL;
struct sidTuneInfo mySidInfo;
struct emuConfig myEmuConfig;
int curSong;
BOOL paused;

void mixer ( void * userdata, u8 * stream, int len )
  {
	  if (paused)
	  	{
		  	int i;
		  	for (i=0;i<len>>1;++i)
		  		((u16*)stream)[i] = 0x8000;
	  	} else
	  	{
    		sidEmuFillBuffer ( *myEmuEngine, *myTune, stream, len );
    	}
  }

static int fsize ( char * fname )
	{
		unsigned long s;
		
		if (GpFileGetSize(fname,&s)!=SM_OK)
			return -1;
		else
			return (int)s;
	}

void PrintSongInfo(void)
	{
		char bla[256];
		int yp;

   	PrintStringCenter ( screenbuf, "SIDPLAY V0.01B", 0, 16 );
   	PrintStringCenter ( screenbuf, "2002,2003 BY CHN", 16, 16 );
   	
   	yp = 80;
   	yp+=PrintString ( screenbuf, mySidInfo.nameString, 0, yp, 16 );
   	yp+=PrintString ( screenbuf, mySidInfo.authorString, 0, yp, 16 );
   	yp+=PrintString ( screenbuf, mySidInfo.copyrightString, 0, yp, 16 );
   	yp+=16;
   	sprintf ( bla, "Subsong: %02d of %02d", curSong, mySidInfo.songs );
   	yp+=PrintString ( screenbuf, bla, 0, yp, 16 );
   	sprintf ( bla, "Free mem: %d", gm_availablesize() );
   	PrintStringCenter ( screenbuf, bla, 240-16, 16 );
	}

/* Sets a single GP32 palette entry */
void GpSetPaletteEntry ( u8 i, u8 r, u8 g, u8 b )
  {
    GP_PALETTEENTRY entry = GP_RGB24(r,g,b);
    GpPaletteEntryChange ( i, 1, &entry, 0 );
  }

void GpMain (void * arg)
  {
    int * t;
    GPSOUNDBUF sb;
    BOOL sid_loaded = FALSE;
    BOOL stereo = TRUE;
    BOOL keypressed = FALSE;
    u8 * sidmem = NULL;
    int sidsize = -1;
    int angle = 0;
    int i;
    
    paused = FALSE;
/*
		GpClockSpeedChange (67500000, 0x25002, 1);
		GpClockSpeedChange (80000000, 0x48012, 1);
		GpClockSpeedChange (99000000, 0x3a002, 1);
		GpClockSpeedChange(110000000, 0x2f011, 1);
		GpClockSpeedChange(120000000, 0x34011, 1);
    GpClockSpeedChange (132000000, 0x3a011, 3);
*/
/*    GpClockSpeedChange (132000000, 0x3a011, 3);*/

    /* Allocate memory for the screen rendering buffer */
    screenbuf = (u8*)malloc(320*240);
    memset((void*)screenbuf, 0, 320*240);

    /* Initialize the LCD */
    GpGraphicModeSet(8, t);
    GpLcdSurfaceGet(&gpDraw[0], 0);
    GpLcdSurfaceGet(&gpDraw[1], 1);
    GpSurfaceSet(&gpDraw[0]);
    GpLcdEnable();
    GpFatInit();
    GpKeyInit();

    GpRelativePathSet ( "gp:\\gpmm\\sid" );

    GpSetPaletteEntry ( 128, 0, 0, 0 );
    GpSetPaletteEntry ( 129, 255, 255, 255 );
    GpSetPaletteEntry ( 130, 255, 0, 0 );
    GpSetPaletteEntry ( 131, 0, 255, 0 );
    GpSetPaletteEntry ( 132, 0, 0, 255 );
    GpSetPaletteEntry ( 133, 255, 255, 0 );
    GpSetPaletteEntry ( 134, 255, 96, 96 );
    
    /* Picture */
    for (i=0;i<16;++i)
    	{
	    	u8 r = ((u16)bgpicpal[(i*3)+0]*24)/8;
	    	u8 g = ((u16)bgpicpal[(i*3)+1]*24)/8;
	    	u8 b = ((u16)bgpicpal[(i*3)+2]*24)/8;
	    	GpSetPaletteEntry ( i, r, g, b );
    	}

    /* Font */
    for (i=0;i<8;i++)
    	{
	    	u8 r = fontpal[(i*3)+0]<<2;
	    	u8 g = fontpal[(i*3)+1]<<2;
	    	u8 b = fontpal[(i*3)+2]<<2;
	    	GpSetPaletteEntry ( i+16, r, g, b );
    	}
    	
    PrintStringCenter ( screenbuf, "Please wait...", 120-8, 16 );
    XlatVGABuffer ( screenbuf, gpDraw[0].ptbuffer );

    myEmuEngine = new emuEngine();
 		myEmuEngine->getConfig ( myEmuConfig );
   	myEmuConfig.frequency = 44100;
 		myEmuConfig.bitsPerSample = SIDEMU_16BIT;
 		myEmuConfig.emulateFilter = true;
 		myEmuConfig.sampleFormat = SIDEMU_UNSIGNED_PCM;

 		sb.freq = stereo ? PCM_S44 : PCM_M44;
 		sb.format = PCM_16BIT;
 		sb.samples = 44100/50;
 		sb.userdata = NULL;
 		sb.callback = mixer;
 		sb.pollfreq = 4*(44100/sb.samples);

    while (1)
    	{
	    	u16 zoomsp=16,scale=0x0100;
				char fname[256];
				char fnamecopy[256];
				BOOL loaded = FALSE;

				while (!loaded)
					{
	    			FSEL fsel;
	    			F_HANDLE fh;
	    			unsigned long tl;
	    			int fs;
	    	
        		fsel.bgcolor = 128;
        		fsel.dircolor = 133;
        		fsel.filecolor = 129;
        		fsel.dirselcolor = 131;
        		fsel.fileselcolor = 131;
        		fsel.curdircolor = 134;
        		fsel.fname = fname;
        		fsel.gpDraw = &gpDraw[0];

						if (sid_loaded)
							{
								fsel.fnamepreselect = fnamecopy;
							} else
							{
								fsel.fnamepreselect = NULL;
							}
				
						if ( FileRequest ( &fsel, NULL, sid_loaded ) == 0 )
							{
								u8 * tempmem;
								fs = fsize ( fname );
								tempmem = (u8*)malloc ( fs );
								GpFileOpen ( fname, OPEN_R, &fh );
								GpFileRead ( fh, tempmem, fs, &tl );
								GpFileClose ( fh );
				
								if (sid_loaded)
									{
										myTune->load ( tempmem, fs );
									} else
									{
    								myTune = new sidTune ( tempmem, fs );
    							}

    						if (!*myTune)
    							{	/* Failed loading the SID */
	    							free ( tempmem );
	    							if (sid_loaded)
	    								{
	    									myTune->load ( sidmem, sidsize );
	    									loaded = TRUE;
    									} else
	    								{
    										loaded = FALSE;
    									}
    							} else
    							{	/* Loaded the SID sucessfully */
										if ( sidmem!=NULL )
											free ( sidmem );

										sidmem = tempmem;
										sidsize = fs;
	    							strcpy ( fnamecopy, fname );
    								sid_loaded = loaded = TRUE;
    							}

    						myTune->getInfo ( mySidInfo );

	    					myEmuConfig.channels = stereo ? SIDEMU_STEREO : SIDEMU_MONO;
    						myEmuEngine->setConfig ( myEmuConfig );

    						sidEmuInitializeSong(*myEmuEngine,*myTune,curSong=1);
    						myEmuEngine->resetSecondsThisSong();
    					} else
    					loaded = TRUE;
    			}

    		GpSoundBufStart(&sb);

    		while ((GpKeyGet()&GPC_VK_SELECT)==0)
    			{
	    			int k = GpKeyGet();
	    			ZoomStr zm;
	    			
						/* Check the keypad */
/*	    			if (k&GPC_VK_FA)
	    				GpAppExit();
	    			else*/
	    			if ((k&GPC_VK_START) && !keypressed)
	    				{
		    				keypressed = TRUE;
		    				paused = !paused;
	    				} else
	    			if ((k&GPC_VK_FL) && !keypressed && stereo)
	    				{
		    				keypressed = TRUE;
		    				stereo = FALSE;
		    				
		    				GpSoundBufStop();
		    				sb.freq = stereo ? PCM_S44 : PCM_M44;
					    	myEmuConfig.channels = stereo ? SIDEMU_STEREO : SIDEMU_MONO;
				    		myEmuEngine->setConfig ( myEmuConfig );
		    				GpSoundBufStart(&sb);
	    				} else
	    			if ((k&GPC_VK_FR) && !keypressed && !stereo)
	    				{
		    				keypressed = TRUE;
		    				stereo = TRUE;
		    				
		    				GpSoundBufStop();
		    				sb.freq = stereo ? PCM_S44 : PCM_M44;
					    	myEmuConfig.channels = stereo ? SIDEMU_STEREO : SIDEMU_MONO;
				    		myEmuEngine->setConfig ( myEmuConfig );
		    				GpSoundBufStart(&sb);
	    				} else
	    			if ((k&GPC_VK_UP) && !keypressed)
	    				{
		    				keypressed = TRUE;
		    				
		    				GpSoundBufStop();
		    				sidEmuInitializeSong(*myEmuEngine,*myTune,curSong);
		    				myEmuEngine->resetSecondsThisSong();
		    				GpSoundBufStart(&sb);
	    				} else
	    			if ((k&GPC_VK_LEFT) && !keypressed)
	    				{
		    				keypressed = TRUE;

		    				if (curSong>1)
		    					{
		    						curSong--;
		    						GpSoundBufStop();
				    				sidEmuInitializeSong(*myEmuEngine,*myTune,curSong);
				    				myEmuEngine->resetSecondsThisSong();
				    				GpSoundBufStart(&sb);
			    				}
	    				} else
	    			if ((k&GPC_VK_RIGHT) && !keypressed)
	    				{
		    				keypressed = TRUE;

		    				if (curSong<mySidInfo.songs)
		    					{
		    						curSong++;
		    						GpSoundBufStop();
				    				sidEmuInitializeSong(*myEmuEngine,*myTune,curSong);
				    				myEmuEngine->resetSecondsThisSong();
				    				GpSoundBufStart(&sb);
			    				}
	    				} else
	    			if (k==0)
	    				{
								keypressed = FALSE;
	    				}

	    			/* Zoom Background */
    				zm.src = bgpicraw;
    				zm.dest = screenbuf;
    				zm.xpos = 0;
    				zm.ypos = 0;
    				zm.xscale = zm.yscale = 0x0008;
	    			if (scale<=0x0010 || scale>=0x0400)
	    				zoomsp = -zoomsp;
    				zm.xscale=zm.yscale=scale;
    				zm.xpos = -(160*zm.xscale) + (sinmul(angle, 0x8000)<<2);
    				zm.ypos = -(120*zm.yscale) + (cosmul(angle, 0x8000)<<2);
    				Zoom ( &zm );
   					angle+=16;
   					scale+=zoomsp;

    				/* Print Song Info */
    				PrintSongInfo();

   					XlatVGABuffer ( screenbuf, gpDraw[0].ptbuffer );
    			}
    		GpSoundBufStop();
    	}
    GpAppExit();
  }
