//
// /home/ms/source/sidplay/libsidplay/fformat/RCS/pp_.cpp,v
//
// For more info get ``depp.cpp'' and ``ppcl.cpp''.

#include "pp_.h"

// exports
bool ppIsCompressed();
udword ppUncompressedLen();  // return the length of the uncompressed data
const char* ppErrorString;

static const char PP_ID[] = "PP20";

static const udword PP_BITS_FAST = 0x09090909;
static const udword PP_BITS_MEDIOCRE = 0x090a0a0a;
static const udword PP_BITS_GOOD = 0x090a0b0b;
static const udword PP_BITS_VERYGOOD = 0x090a0c0c;
static const udword PP_BITS_BEST = 0x090a0c0d;

static const char text_packeddatacorrupt[]	= "PowerPacker: Packed data is corrupt";
static const char text_unrecognized[]	= "PowerPacker: Unrecognized compression method";
static const char text_uncompressed[] = "Not compressed with PowerPacker (PP20)";
static const char text_notenoughmemory[] = "Not enough free memory";
static const char text_fast[] = "PowerPacker: fast compression";
static const char text_mediocre[] = "PowerPacker: mediocre compression";
static const char text_good[] = "PowerPacker: good compression";
static const char text_verygood[] = "PowerPacker: very good compression";
static const char text_best[] = "PowerPacker: best compression";

static ubyte* sourceBuf;
static ubyte* readPtr;
static ubyte* writePtr;
static ubyte* startPtr;
static udword current;       // compressed data longword
static int bits;             // number of bits in 'current' to evaluate
static ubyte efficiency[4];
static udword outputLen;
static bool isCompressed;
static bool globalError;


// Move four bytes to Motorola big-endian double-word.
static inline void bytesTOudword()
{
	readPtr -= 4;
	if ( readPtr < sourceBuf )
	{
		ppErrorString = text_packeddatacorrupt;
		globalError = true;
	}
	else 
		current = readEndian(*readPtr,*(readPtr+1),*(readPtr+2),*(readPtr+3));
}

static void ppFreeMem()
{
	if (sourceBuf != 0)      // should not be required
		delete[] sourceBuf;
	sourceBuf = 0;
}

static inline udword ppRead(int count)
{
	udword data = 0;
	// read 'count' bits of packed data
	for (; count > 0; count--)
	{
		// equal to shift left
		data += data;
		// merge bit 0
		data = (data | (current&1));
		current = (current >> 1);
		if ((--bits) == 0)
		{
			bytesTOudword();
			bits = 32;
		}
	}
	return data;
}

static inline void ppBytes()
{
	udword count, add;
	count = (add = ppRead(2));
	while (add == 3)
	{
		add = ppRead(2);
		count += add;
	}
    count++;
	for (; count > 0 ; count--)
	{
		if (writePtr > startPtr)
		{
			*(--writePtr) = (ubyte)ppRead(8);
		}
		else
		{
			ppErrorString = text_packeddatacorrupt;
			globalError = true;
		}
	}
}

static inline void ppSequence()
{
	udword offset, length, add;
	int offsetBitLen;
	length = ppRead( 2 );  // length -2
	offsetBitLen = (int)efficiency[length];
	length += 2;
	if ( length != 5 )
		offset = ppRead( offsetBitLen );
	else
	{
		if ( ppRead( 1 ) == 0 )
			offsetBitLen = 7;
		offset = ppRead( offsetBitLen );
		add = ppRead( 3 );
		length += add;
		while ( add == 7 )
		{
			add = ppRead(3);
			length += add;
		}
	}
	for ( ; length > 0 ; length-- )
	{
		if ( writePtr > startPtr )
		{
			--writePtr;
			*writePtr = *(writePtr+1+offset);
		}
		else
		{
			ppErrorString = text_packeddatacorrupt;
			globalError = true;
		}
	}
}

udword ppUncompressedLen()
{
	return outputLen;
}

bool ppIsCompressed()
{
	return isCompressed;
}

